# -*- coding: UTF-8 -*- 
#   universal_cam.py  2020.7.25   kita

#!! Universal_cam クラス本体

#履歴   
#   2021.5.20   cam_no 34   非対称変形台形曲線を追加
#               cam_no 102  非対称等加速度曲線を追加、get_s, get_v(), get_ratio追加 
#   2021.6.10   get_t() 追加
#   2021.6.27   zenkin(func, t00, t01, dt, target_val, gosa0) 追加
#   2021.8.20   get_05Tmirror(self): V:0.5TY軸反転カム取得メソッド
#               sva_nn_real にX軸反転、Y軸反転カム諸元の取得モードを追加  menu == 5 参照　


import sys
import numpy as np
try:
    import matplotlib.pyplot as plt
except:
    print('matplotlib がインストールされていません。グラフの作図機能は使えません')

"""
__init__(self,cam_no,z,*tt):
cam_no      22: Cycroidal
            25:標準変形台形         26:標準変形正弦        27:標準変形等速度
            45:片停留変形台形(m=1)  48:片停留変形正弦
            51:無停留変形台形       52:無停留変形等速度
            1:早送り(v7!=0)         2:切削送り(等速->停止)
            0:t0～t7を指定する	    3:全ての変数を指定
            1: z   早送り曲線のパラメータ　t3 = 1/(1+z)
            0:*tt   t0-t7 をListで指定する場合に使用する
            34:非対称変形台形(t1 = z)
 
sva(self,t):  時間t(0-1)の時の(s, v, a)を返す。

sva_nn(self,nn,mode):   t=0-1を nn 分割した(s, v, a)を返す。mode=1 でPlot out


"""

class Universal_cam:
    # universal_camの初期化  インスタンス（実体）作成時に実行される
    def __init__(self, cam_no, z=0.9, ratio=1.0, *tt):
        s0 = 0; s7 = 1; v0 = 0; v7 = 0; pi = np.pi #; ratio = 1
        if cam_no == 22:    # Cycroidal
            tt = [0.0,1/4,1/4,1/2,1/2,3/4,3/4,1.0]; n_a0 = 2; ttl = 'Cycroidal'
        elif cam_no == 25:    #　標準変形台形
            tt = [0.0,1/8,3/8,1/2,1/2,5/8,7/8,1.0]; n_a0 = 2; ttl = 'Std_Modified_Trapezoid' #'標準変形台形'
        elif cam_no == 26:    #　標準変形正弦
            tt = [0.0,1/8,1/8,1/2,1/2,7/8,7/8,1.0]; n_a0 = 2; ttl = 'Std_Modified_Sine'    #'標準変形正弦'
        elif cam_no == 27:    #　標準変形等速度
            tt = [0.0,1/16,1/16,1/4,3/4,15/16,15/16,1.0]; n_a0 = 2; ttl = 'Std_Mod_Const_Velocity'   #標準変形等速度'
        elif cam_no == 45:    #　片停留変形台形(V6=V7=0, m=ratio=1) 
            alf = 1/3; t0=0.0; t3=t4=0.527565; t1=alf*t3; t2=t3*(1-alf); t5=t3+0.2; t6=t7=1.0 # m=1.0000 Amp=4.689
            #alf = 1/3; t0=0.0; t3=t4=4/8; t1=alf*t3; t2=t3*(1-alf); t5=t3+0.13; t6=t7=1.0 # m=0.84 Amp=5.05 
            #alf = 1/3; t0=0.0; t3=t4=3/8; t1=alf*t3; t2=t3*(1-alf); t5=3.5/8; t6=t7=1.0 # m=0.49 Amp=6.885
            #tt = [0.0,1/8,3/8,1/2,1/2,5/8,1.0,1.0]; ratio = 0.8; n_a0 = 1; ttl = 'One_Dwell_Mod_Trapezoid'  #片停留変形台形'
            c13 = t3*(4*alf/pi+1-2*alf); c57 = t5*(2/pi-1)+(1-2*t3/pi)
            ratio = c13/c57; n_a0 = 1 
            tt = [t0,t1,t2,t3,t4,t5,t6,t7]; ttl = 'One_Dwell_Mod_Trapezoid'      
        elif cam_no == 48:    #　片停留変形正弦
            tt = [0.0,1/8,1/8,4/8,4/8,1.0,1.0,1.0]; n_a0 = 1; ttl = 'One_Dwell_Mod_Sine'  #片停留変形正弦'
        elif cam_no == 51:    #　無停留変形台形
            tt = [0.0,0.0,2/8,4/8,4/8,6/8,1.0,1.0]; n_a0 = 0; ttl = 'No_Dwell_Mod_Trapezoid'  #無停留変形台形'
        elif cam_no == 52:    #　無停留変形等速度
            tt = [0.0,0,0,1/8,7/8,1,1.0,1.0]; n_a0 = 0; ttl = 'No_Dwell_Mod_Cst_Velosity'  #無停留変形等速度'
            #tt = [0.0,0,2/8,4/8,4/8,6/8,1.0,1.0]; n_a0 = 0; ttl = 'No_Dwell_Mod_Cst_Velosity'  #無停留変形等速度'
        elif cam_no == 1:     #  早送り(v7!=0)
            t0 = 0; t3 = 1/(1+z); t1 = t3/4; t2 = t1; t7 = 1; t5 = t7-(t7-t3)/4; t4 = t3; t6 = t5; n_a0 = 2;
            tt = [t0,t1,t2,t3,t4,t5,t6,t7]; ttl = 'hayaokuri(z=%1.1f)' % z     
        elif cam_no == 2:     #  片停留切削送り(等速->停止)
            t0 = 0; t1 = 0; t2 = 0; t3 = 0; t4 = z; t5 = 1; t6 = 1; t7 = 1; s0 = 0; s7 = 1; v7 = 0; n_a0 = 1
            self.a_mp=a_mp = 0;ratio = 0; self.a_mm=a_mm = pi/(4*(1-t4)**2/pi + 2*(1-t4)*t4); v0 = 2*(1-t4)*a_mm/pi
            tt = [t0,t1,t2,t3,t4,t5,t6,t7]; ttl = 'katateiryu_sessakuokuri'   #片停留切削送り(等速->停止)' 
        elif cam_no == 0:     #　t0～t7を指定する
            n_a0 = 0; v0 = z
            if v0 == 0:
                ttl = 't0-t7 settei' 
            else:
                ttl = 'v0 t0-t7 settei' 
                
        elif cam_no == 34:  # 非対称両停留変形台形  J3=J4  引数 (z=t1, ratio= Amm/Amp = m)
            t1 = z; m = ratio
            t3 = ((1-m*m)*(2-4/pi)*t1 + m)/(m+1)    # J3=J4 & v7=0

            t0 = 0; t2 = t3 - t1; t4 = t3; t5 = t3 + m*t1; t7 = 1; t6 = t7 - m*t1; 
            tt = [t0,t1,t2,t3,t4,t5,t6,t7]
            n_a0 = 2; ttl = 'Asym_Modified_Trapezoid(J3=J4)' # 非対称変形台形

        # Cma No 34 より降格
        elif cam_no == 340:  # 非対称変形台形  J3!=J4  引数 (z=t1, ratio= Amm/Amp = m)
            t1 = z
            #t1 = 1/16
            #t1 = 1/8
            t3 = (ratio+2*t1*(1-ratio)-4*t1/pi*(1-ratio))/(1+ratio)
            t0 = 0; t2 = t3 - t1; t4 = t3; t5 = t3 + t1; t7 = 1; t6 = t7 - t1; 
            tt = [t0,t1,t2,t3,t4,t5,t6,t7]
            n_a0 = 2; ttl = 'Asym_Modified_Trapezoid' #'非対称変形台形'

        #elif cam_no == 100: # V0指定 z を利用
        #    v0 = z
        #    #tt = [t0,t1,t2,t3,t4,t5,t6,t7]
        #    n_a0 = 0; ttl = 'V0 t0-t7 settei' # V0 != 0
        elif cam_no == 102: # 非対称等加速度曲線
            t3 = ratio/(1+ratio)
            t0 = t1 = 0; t2 = t4 = t5 = t3; t6 = t7 = 1 
            tt = [t0,t1,t2,t3,t4,t5,t6,t7]
            n_a0 = 0; ttl = 'Asym_Constant_Accel' #'非対称等加速度'
            

        #print(tt)
        self.tt = tt; self.ttl = ttl; self.cam_no = cam_no; self.z = z;
        self.s0 = s0;  self.s7 = s7;  self.v0 = v0; 
        self.ratio=ratio; self.n_a0 = n_a0;
    
        #print(tt)
        #self.t*=t0～t7を指定する
        #print(len(tt))
        self.t0=t0 = tt[0]; self.t1=t1 = tt[1]; self.t2=t2 = tt[2]; self.t3=t3 = tt[3]
        self.t4=t4 = tt[4]; self.t5=t5 = tt[5]; self.t6=t6 = tt[6]; self.t7=t7 = tt[7]
        self.c0=c0 = 0
        self.c1=c1 = 2*(t1-t0)/pi
        self.c2=c2 = t2-t1
        self.c3=c3 = 2*(t3-t2)/pi
        self.c4=c4 = t4-t3
        self.c5=c5 = 2*(t5-t4)/pi
        self.c6=c6 = t6-t5
        self.c7=c7 = 2*(t7-t6)/pi
  
        c_tmp1 = s7-s0-self.v0*(t7-t0)
        c_tmp2 = c7**2+0.5*(c6**2)-(c5**2)+c6*(t7-t6)+c5*(t7-t4)
        c_tmp3 = c3**2+0.5*(c2**2)-(c1**2)+c3*(t7-t3)+c2*(t7-t2)+c1*(t7-t0)
        tmp = -ratio*c_tmp2+c_tmp3
        #print('universal_cam', tmp, c_tmp1, c_tmp2, c_tmp3)
        
        if tmp != 0:
            self.a_mp=a_mp = c_tmp1/(-ratio*c_tmp2+c_tmp3)
            #print('Universal_a_mp = %3.3f' % a_mp)
        elif cam_no != 2:
            print('Ampが計算できませんでした')
            sys.exit()
        if  cam_no != 2:
            self.a_mm=a_mm = ratio*a_mp
     
        self.v1=v1 = c1*a_mp + v0
        self.v2=v2 = c2*a_mp + v1
        self.v3=v3 = c3*a_mp + v2
        #print('Universal_v3 = %3.3f' % v3)
        self.v4=v4 = v3
        self.v5=v5 = -c5*a_mm + v4
        self.v6=v6 = -a_mm*c6 + v5
        self.v7=v7 = -c7*a_mm + v6
        #print(v7)

        self.s1=s1 = -(c1**2)*a_mp + v1*(t1-t0) + s0
        self.s2=s2 = a_mp/2*c2*c2 + v1*c2 + s1
        self.s3=s3 = c3*c3*a_mp + v2*(t3-t2) + s2
        self.s4=s4 = v3*c4 + s3
        self.s5=s5 = c5*c5*a_mm+v5*(t5-t4) + s4
        self.s6=s6 = -a_mm/2*c6*c6 + v5*c6 + s5
        self.s7=s7 = -c7*c7*a_mm + v6*(t7-t6) + s6
    
    #---------------------------------------------------------------
    #    t=0-1 の値を与えると s:Stroke v:速度 a:加速度 を返す   
    #---------------------------------------------------------------
    def sva(self,t):
        #print(t,self.t1,self.a_mp)
        if(self.t0 <= t < self.t1):
            p = (t-self.t0)/self.c1
            #p = p_rad*180/piuniversal_cam.py
            a = self.a_mp*np.sin(p)
            j = self.a_mp/self.c1*np.cos(p)
            v = self.c1*self.a_mp*(1-np.cos(p))+self.v0
            s = -(self.c1**2)*self.a_mp*np.sin(p) + self.v1*(t-self.t0) + self.s0
        elif(self.t1 <= t <= self.t2):
        #elif(self.t1 <= t < self.t2):
            p = t-self.t1
            a = self.a_mp; j = 0
            v = self.a_mp*p + self.v1
            s = self.a_mp/2*p*p + self.v1*p + self.s1
        elif(self.t2 < t < self.t3):
        #elif(self.t2 <= t < self.t3):
            p = (t-self.t2)/self.c3
            #p = p_rad*180/pi
            a = self.a_mp*np.cos(p)   
            j = -self.a_mp/self.c3*np.sin(p)
            v = self.c3*self.a_mp*np.sin(p) + self.v2

            v_m = self.v3

            s = -(self.c3**2)*self.a_mp*(np.cos(p)-1) + self.v2*(t-self.t2) + self.s2
        elif(self.t3 <= t < self.t4):
            a = 0; j = 0; v = self.v3
            s = self.v3*(t-self.t3) + self.s3
        elif(self.t4 <= t < self.t5):
            p = (t-self.t4)/self.c5
            #p = p_rad*180/pi
            a = -self.a_mm*np.sin(p)
            j = -self.a_mm/self.c5*np.cos(p)
            v = self.c5*self.a_mm*(np.cos(p)-1) + self.v4
            s = self.c5**2*self.a_mm*np.sin(p) + self.v5*(t-self.t4) + self.s4
        elif(self.t5 <= t < self.t6):
            p = t-self.t5
            a = -self.a_mm; j=0
            v = -self.a_mm*p + self.v5
            s = -(self.a_mm/2.0*(p**2)) + self.v5*p + self.s5
        #elif(self.t6 <= t <= self.t7):
        elif(self.t6 <= t):
            if(self.c7 == 0):
                p=0; a = -self.a_mm; j = 0
            else:
                p = (t-self.t6)/self.c7
                #p = p_rad*180/pi
                a = -self.a_mm*np.cos(p)
                j = self.a_mm/self.c7*np.sin(p)
            v = -self.c7*self.a_mm*np.sin(p) + self.v6
            s = self.c7**2*self.a_mm*(np.cos(p)-1) + self.v6*(t-self.t6) + self.s6
        """
        elif(self.t7 <= t):
            if(self.n_a0 == 2):      # 両停留
                a = 0; j = 0; v = self.v7
                #s = s7
                s = self.s7+self.v7*(t-self.t7)
            elif(self.n_a0 == 1): # 片停留 
                a = -self.a_mm; j = 0; s = 1; v = self.v7
            elif(self.n_a0 == 0):    # 無停留 */
                a = -self.a_mm; j = 0; v = self.v7
                s = self.s7+self.v7*(t-self.t7)
        """   
            #print('t=',t,'  s=',s,'  a=',a)
        
        #print('S= %2.5f  V= %2.5f  A= %2.5f' % (s,v,a))

        return (s, v, a)
        #return (round(s,10), round(v,10), round(a,10))      # 少数点以下10桁で丸める

    def sva_graph(self, sva_all, max_x, scl_s = 1.0):    ### sva全リスト、x軸最大値
        nn = len(sva_all[0])
        for i in range(nn):
            sva_all[0][i] = sva_all[0][i] * scl_s
            
        max_s = max(sva_all[0])/scl_s; min_s = min(sva_all[0])/scl_s 
        max_v = max(sva_all[1]); min_v = min(sva_all[1])
        max_a = max(sva_all[2]); min_a = min(sva_all[2])
        dx = max_x/(nn-1)
        #dx = max_x/(nn)
        start = 0; end = max_x
        x = np.arange(start,end+dx,dx)  # 0から1まで0.1刻みで生成
        #x = np.arange(start,end,dx)  # 0から1まで0.1刻みで生成
        if len(x) > nn:
            x = np.arange(start,end-dx,dx)  # 0から1まで0.1刻みで生成
        if len(x) < nn:
            x = np.arange(start,end+dx,dx)        
        
        
        plt.grid() # グリッド線
        # 0線
        #p = plt.hlines([0], min_s, max_s, "blue", linestyles='dashed')     # hlines
        #p = plt.hlines([0], start, end, "gray", linestyles='dashed')     # hlines

        if max_x == 1:      #無次元と判断（たまたまt_h = 1の場合も有り得るが)
            plt.plot(x,sva_all[0],label="Stroke   %3.3f : %3.3f" % (max_s, min_s),color='g')
            plt.plot(x,sva_all[1],label="Velocity %3.3f : %3.3f" % (max_v, min_v),color='b')
            plt.plot(x,sva_all[2],label="Accel    %3.3f : %3.3f" % (max_a, min_a),color='r')
            plt.title('%s(NoneDim)' % self.ttl)
        else:    
            plt.plot(x,sva_all[0],label="Stroke   %3.3f : %3.3f[rad]" % (max_s, min_s),color='g')
            plt.plot(x,sva_all[1],label="Velocity %3.3f : %3.3f[rad/sec]" % (max_v, min_v),color='b')
            plt.plot(x,sva_all[2],label="Accel    %3.3f : %3.3f[rad/sec2]" % (max_a, min_a),color='r')
            plt.title(self.ttl)

        plt.legend()
        plt.show() 
        #plt.show(block=False)   # グラフを閉じなくても復帰

    def graph_close(self):
        plt.close()


    #---------------------------------------------------------------------------------------------
    #   s(変位) v(速度) a(加速度) の全ての配列を返す(無次元)。nn:配列サイズ(Ex.100)  mode = 1でグラフ作図 
    #---------------------------------------------------------------------------------------------
    def sva_nn(self, nn, mode, max_x=1.0, scl_s=1.0): # mode == 1 グラフ作図(無次元)
        sva_all = [[],[],[]]   # s,v,a の配列(リスト) 
        tmp=[]
        #cam = Universal_cam(self.cam_no,self.z, self.ratio, *self.tt)
        for i in range(nn+1):  # i = 0 - 100
            t = i/nn
            tmp = self.sva(t)
            #print(tmp)
            sva_all[0].append(tmp[0])
            sva_all[1].append(tmp[1])
            sva_all[2].append(tmp[2])
        
        if mode == 1:
            #max_x = 1
            self.sva_graph(sva_all,max_x, scl_s)
            #print(self.sva_all[0])

        return  sva_all


    #---------------------------------------------------------------------------------------------
    #   切断線 0<tx<1 に割り付けた、s(変位) v(速度) a(加速度) の全ての配列を返す(無次元)。 
    ###---------------------------------------------------------------------------------------------
    def sva_nn_tx(self, nn, mode, tx, max_x=1.0, scl_s=1.0): # mode == 1 グラフ作図(無次元)
        sva_all = [[],[],[]]   # s,v,a の配列(リスト) 
        tmp=[]; dt = tx/nn
        #cam = Universal_cam(self.cam_no, self.z, self.ratio, *self.tt)
        for i in range(nn+1):  # i = 0 - 100
            t = dt*i
            tmp = self.sva(t)
            #print(tmp)
            sva_all[0].append(tmp[0])
            sva_all[1].append(tmp[1])
            sva_all[2].append(tmp[2])
            #sva_all[1].append(tmp[1]/tx)
            #sva_all[2].append(tmp[2]/tx**2)
        
        if mode == 1:
            #max_x = 1
            cam.sva_graph(sva_all, max_x, scl_s)
            #print(self.sva_all[0])

        return  sva_all


    #---------------------------------------------------------------------------------------------------
    #   s(変位) v(速度) a(加速度) の全ての配列を返す(有次元)。nn:配列サイズ(Ex.100)  mode = 1でグラフ作図 
    #       2021.8.20   sxy:'s' : 標準Cam  'x' : X軸反転形Cam 　'y':Y軸反転Cam(後方接続)
    #                   s00: 初期ストローク
    #---------------------------------------------------------------------------------------------------
    def sva_nn_real(self, nn, mode, h, t_h, sxy = 's', s00=0):
        #nn:配列サイズ  mode:1でグラフ作図  h:Stroke  t_h:実時間
        #sxy:'s':標準Cam 'x':X軸反転形Cam　'y':Y軸反転Cam(後方接続)  s00:初期Stroke

        sva_all_real = [[],[],[]]
        #sva_all = self.sva_nn(nn, 0)
        if sxy == 's':
            #cam = Universal_cam(self.cam_no,self.z, self.ratio, *self.tt)
            for i in range(nn+1):  # i = 0 - 100
                t = i/nn
                tmp = self.sva(t)
                sva_all_real[0].append(tmp[0]*h+s00)
                sva_all_real[1].append(tmp[1]*h/t_h)
                sva_all_real[2].append(tmp[2]*h/t_h**2)

        if sxy == 'x':  # h=-h と同じことだから不要かな?
            #cam = Universal_cam(self.cam_no,self.z, self.ratio, *self.tt)
            for i in range(nn+1):  # i = 0 - 100
                t = i/nn
                tmp = self.sva(t)
                sva_all_real[0].append(-(tmp[0]*h+s00))
                sva_all_real[1].append(-tmp[1]*h/t_h)
                sva_all_real[2].append(-tmp[2]*h/t_h**2)

        elif sxy == 'y':
            for i in range(nn+1):  # i = 0 - 100
                t = i/nn
                tmp = self.sva(1-t)
                sva_all_real[0].append(tmp[0]*h+s00)
                sva_all_real[1].append(-tmp[1]*h/t_h)
                sva_all_real[2].append(tmp[2]*h/t_h**2) #2022.10.20修正
                #sva_all_real[2].append(-tmp[2]*h/t_h**2)

        elif sxy == 'xy':
            for i in range(nn+1):  # i = 0 - 100
                t = i/nn;
                tmp = self.sva(1-t)
                sva_all_real[0].append(-tmp[0]*h+s00)
                sva_all_real[1].append(tmp[1]*h/t_h)
                sva_all_real[2].append(-tmp[2]*h/t_h**2) 
        
        if mode == 1:
            self.sva_graph(sva_all_real,t_h)
            
        return sva_all_real
 
    #--------------------------------------------------------------------
    #   0.5T軸反転形カム(mirror)を返す関数 cam_m = cam.get_05Tmirror()
    #--------------------------------------------------------------------
    def get_05Tmirror(self):
        t0=1-self.t7; t1=1-self.t6; t2=1-self.t5; t3=1-self.t4
        t4=1-self.t3; t5=1-self.t2; t6=1-self.t1; t7=1-self.t0
        #t0=1-self.cam.t7; t1=1-self.cam.t6; t2=1-self.cam.t5; t3=1-self.cam.t4
        #t4=1-self.cam.t3; t5=1-self.cam.t2; t6=1-self.cam.t1; t7=1-self.cam.t0
        #t0=1-cam.t7; t1=1-cam.t6; t2=1-cam.t5; t3=1-cam.t4
        #t4=1-cam.t3; t5=1-cam.t2; t6=1-cam.t1; t7=1-cam.t0
        ttm = [t0,t1,t2,t3,t4,t5,t6,t7]; rat = 1/self.ratio
        #print(ttm)
        cam_05Tm = Universal_cam(0, 0, rat, *ttm)

        return cam_05Tm



    #-----------------------------------------------------
    #   s(変位) からt(無次元時間)を逆算する
    #-----------------------------------------------------

    def s2t(self, target_s, gosa0=0.0000001):
        if target_s > 1.0:
            print('Target_sが不正です')
            return None

        #cam = Universal_cam(self.cam_no, self.z, self.ratio, *self.tt)
        #gosa0 = 0.0001
        t0 = 0.0 ; t1 = 1.0; dt = 0.1   #初期値
        def zenkin1(t0, t1, dt):
            nn = round((t1-t0)/dt)+1    # int ではなく　四捨五入のround
            lst = [t0+i*dt for i in range(nn)] # t=t0; t<=t1; t+=dtと同じ
            #print(lst)
            for t in lst:
                (s, v, a) = self.sva(t)
                gosa = abs(target_s - s)
                #print(dt, t, s, (target_s - s))
                if t == t0:
                    pre_pm = (target_s - s)/abs(target_s - s)
                    #print('pre_pm= %d' % pre_pm)
                    continue
                if gosa < gosa0:
                    break
                pm = (target_s - s)/gosa

                if pm != pre_pm:    # 符号が反転
                    #print('反転しました')
                    break   # for loop end
                         
            return t, gosa

        while True:
            (tmp_t, tmp_gosa) = zenkin1(t0, t1, dt)
            #print('誤差= %1.8f' % tmp_gosa)
            if tmp_gosa < gosa0:    # 目標達成!
                break
            else:       # まだ誤差が大きい
                t0 = tmp_t - dt
                t1 = tmp_t
                dt = dt/10
                
        return tmp_t

    #   t0-t7を返す (t0, t1, t2, t3, t4, t4, t5, t6, t7) = cam.get_t
    def get_t(self):
        return (self.t0, self.t1, self.t2, self.t3, self.t4, self.t5, self.t6, self.t7)

    #   s0-s7を返す (s0, s1, s2, s3, s4, s4, s5, s6, s7) = cam.get_s
    def get_s(self):
        return (self.s0, self.s1, self.s2, self.s3, self.s4, self.s5, self.s6, self.s7)

    #   v0-v7を返す (v0, v1, v2, v3, v4, v5, v6, v7) = cam.get_v()
    def get_v(self):
        return (self.v0, self.v1, self.v2, self.v3, self.v4, self.v5, self.v6, self.v7)

    #   a0-a7を返す (a0, a1, a2, a3, a4, a5, a6, a7) = cam.get_a()
    def get_a(self):
        return (self.a0, self.a1, self.a2, self.a3, self.a4, self.a5, self.a6, self.a7)

    #   c0-c7を返す (c0, c1, c2, c3, c4, c5, c6, c7) = cam.get_c()
    def get_c(self):
        return (self.c0, self.c1, self.c2, self.c3, self.c4, self.c5, self.c6, self.c7)

    #   ratio, a_mp, a_mmを返す
    def get_ratio(self):
        return (self.ratio, self.a_mp, self.a_mm)


#------------------------------------------------
#   以上がUniversal_cam Classのメンバーメソッド
#   以下、関係関数
#------------------------------------------------



# 漸近法を用いてfuncの戻り値とtarget_valとの誤差がgosa0以下になるような t を返す
def zenkin(func, t00, t01, dt, target_val, gosa0):
    #print('zenkin t00= %2.5f t01= %2.5f ' % (t00, t01))    #!!!
    if t00 > t01:
        print('\n   zenkin  t00= %2.5f   t01= 2.5f  範囲が不正です' % (t00, t01))
        return None

    def for_loop(t00, t01, dt, target_val):
        nn = int(round((t01-t00)/dt))
        lst = [t00+i*dt for i in range(nn)]
        gosa_lst = []
        i = 0; t = 0.0
        for t in lst:
            #print('zenkin t00= %2.5f t01= %2.5f t= %2.5f' % (t00, t01, t))    #!!!
            tmp = func(t)
            #print(t, tmp)
            if tmp == None:
                continue

            tmp_gosa = target_val-tmp
            gosa = abs(tmp_gosa)
            if gosa == 0.0:
                gosa_lst.append(0.0)
                break
            pm = tmp_gosa/gosa
            if i == 0:  pre_pm = pm
            gosa_lst.append(abs(target_val-tmp))
            if pm != pre_pm:    # 符号が反転
                break
            if t == t01:
                return (None, None)
            i += 1
        

        if len(gosa_lst) == 0:
            #print('gosa_lst = 0 t= %f' % t) #!!!
            return (t,999)
        else:
            #print('min(gosa_lst)= %f t01= %f\007' % (min(gosa_lst), t01))    #!!!
            #print(gosa_lst)
            
            return (t, min(gosa_lst))

    while(True):
        (t, gosa_min) = for_loop(t00, t01, dt, target_val)
        if t == None:   # t01までの検索終了
            #print('zenkin 目標値に未到着  t00= %f  t01= %f  dt= %f' % (t00, t01, dt))   #!!!
            return None
        #print(gosa_min)
        if gosa_min < gosa0:    # 目標達成!
            #print('v : zenkin gosa_min= %f gosa0=%f  目標達成!\n ' % (gosa_min, gosa0)) #!!!
            break
        else:       # まだ誤差が大きい
            t00 = t - dt
            t01 = t
            dt = dt/10
            #print('zenkin t00= %f  t01= %f  dt= %f' % (t00, t01, dt))   #!!! 
        if dt == 0.0:
            print('\n    設定値に無理があります(zenkin  dt = 0)\007\n')
            #return t
            return None
    #print('t= %f  gosa_min= %f' % (t, gosa_min))  #!!!
    return t



#------------------------------------
#   0.5TY軸 反転形カムのttm(mirror)を返す関数
#------------------------------------
def get_ttm(tt):
    t0=1-tt[7]; t1=1-tt[6]; t2=1-tt[5]; t3=1-tt[4]
    t4=1-tt[3]; t5=1-tt[2]; t6=1-tt[1]; t7=1-tt[0]
    ttm = [t0,t1,t2,t3,t4,t5,t6,t7]
    return ttm

#-----------------------------------------------------------------------------------
# 漸近法を用いてfuncの戻り値とtarget_valとの誤差がgosa0以下になるような cam を返す
# finc(t):  return (cam, val) このcamがzenkin_cam の戻り値となる 
#-----------------------------------------------------------------------------------
def zenkin_cam(func, t00, t01, dt, target_val, gosa0):
    def for_loop(t00, t01, dt, target_val):
        global cam
        nn = round((t01-t00)/dt)+1
        lst = [t00+i*dt for i in range(nn)]
        gosa_lst = []
        i = 0
        for t in lst:
            (cam, val) = func(t) #この cam が戻り値
            #print(type(cam))
            if val == None:
                continue

            tmp_gosa = target_val-val
            gosa = abs(tmp_gosa)
            pm = tmp_gosa/gosa
            if i == 0:  pre_pm = pm
            gosa_lst.append(abs(target_val-val))
            if pm != pre_pm:    # 符号が反転
                break
            i += 1
            
        if len(gosa_lst) == 0:
            #print('誤差リストが0です')
            return (t,999)
        else:
            #print('min(gosa_lst= %f \007\n' % min(gosa_lst))    #!!!
            return (t, min(gosa_lst))

    while(True):    # gosaが小さくなるまでloop
        (t, gosa_min) = for_loop(t00, t01, dt, target_val)
        #print(gosa_min)
        if gosa_min < gosa0:    # 目標達成!
            print('目標達成!')
            break
        else:       # まだ誤差が大きい
            t00 = t - dt
            t01 = t
            dt = dt/10
        if dt == 0.0:
            print('\n    設定値に無理があります(zenkin_cam  dt = 0)\007\n')
            return None
    return cam



#-----------------------------------------------
#   メインルーチン
#-----------------------------------------------
if __name__ == "__main__":
    import sys

    if len(sys.argv) < 2:   #引き数無し argv
        menu = 1
    else:    # 引数１個 
        menu = int(sys.argv[1])


    if menu == 1:    # Cam No
        cam_no = int(input('Key In Cam No = '))
        cam = Universal_cam(cam_no) 
        all_sva = cam.sva_nn(100,1) 
        print('\n   cam_no= %d  Amp= %f  Amm= %f' % (cam_no, cam.a_mp, cam.a_mm)) 
        t = input('\n   key in   0 < t <=1 =  ')
        sva = cam.sva(float(t))
        print('   s= %10.5f  v= %10.5f  a= %10.5f' % (sva[0],sva[1], sva[2]))
        #cam.graph_close()


    if menu == 17:  # 非対称両停留変形台形カム曲線 m=Amm/Amp 指定 J3 = J4
                    # test_universal_tpcam.py に移動
        #for m in np.arange(0.3, 6, 0.7):
        for m in [0.3, 0.5, 1.0, 2,3,4,5]:
            t35 = -0.014*m*m + 0.14*m - 0.001
            t1 = t35/m
            cam = Universal_cam(34, t1, m)  #非対称両留変形台形
            t23 = cam.t3 - cam.t2
            print('m= %3.1f t1= %5.6f  t35= %5.3f  t35/t23= %5.3f' % (m, t1, t35, t35/t23))
            #print(cam.t2, cam.t3, cam.t5)
            all_sva = cam.sva_nn_real(100, 1, 1.0, 1.0, 's', 0) #非対称変形台形カム曲線
            
            

    if menu == 16:  # 非対称変形台形カム曲線 m=Amm/Amp 指定 J3 = J4
        #ss_m = input('m=Amm/Ampを入力してください    ')
        #m = float(ss_m)
        t35 = 0.125     # m=1
        t35 = 0.15      # m=1.5
        t35 = 0.2       # m=2
        t35 = 0.18      # m=2

        for m in np.arange(0.3, 1, 0.1):
            if m <= 0.5:
                t35 = 0.04
            else:
                t35 = (0.125-1/18) + m/18
            #t35 = .025*m*m + 0.1
            t1 = t35/m
            print('m= %3.1f t1= %5.6f  t35= %5.3f' % (m, t1, t35))
            cam = Universal_cam(34, t1, m)  #片停留変形台形(m=0.8)
            all_sva = cam.sva_nn_real(100, 1, 1.0, 1.0) #非対称変形台形カム曲線


    if menu == 15:  # 非対称変形台形カム曲線 t1, m=Amm/Amp 指定  J3 = J4
        ss_t1, ss_m = input('t1 と m=Amm/Ampを入力してください    ').split()
        t1 = float(ss_t1); m = float(ss_m)
        cam = Universal_cam(34, t1, m)  #片停留変形台形(m=0.8)
        all_sva = cam.sva_nn_real(100, 1, 1.0, 1.0)        # 非対称変形台形カム曲線
        #all_sva = cam.sva_nn_real(100,1, 10.0, 3.0)        # 


    if menu ==14:   ### 切断線の検索
        tx0=0.1; tx1=0.9; t3=0.5; t7=1
        alf=1/4; rat=1; t0=0; t3=0.5; t1=t3*alf; t2=t3*(1-alf); t4=t3
        t5=t3*(rat*alf+1); t6=t7-(t5-t4); 
        nx = 8; dt = (tx1-tx0)/nx; nn = 100

    if menu == 13:  ### sva_nn_tx(nn, mode, tx, scl_s)  切断線のテスト
        tx0=0.1; tx1=0.9; t7=1
        alf=1/4; rat=1; t0=0; t3=0.5; t1=t3*alf; t2=t3*(1-alf); t4=t3
        t5=t3*(rat*alf+1); t6=t7-(t5-t4); 
        nx = 8; dt = (tx1-tx0)/nx; nn = 100

        tt = [t0,t1,t2,t3,t4,t5,t6,t7]  
        cam = Universal_cam(0, 0.0 ,rat, *tt)
                       # nx, mode, tx, scl_x, max_x 
        all_sva0 = cam.sva_nn(nn, 1, t7, 1)
        input('Key In !')
        cam.graph_close()
        
        for i in range(nx):
            tx = (i+1)*dt
            #m = Universal_cam(0, 0.0 ,rat, *tt)
                           # nn, mode, tx, scl_x, max_x 
            all_sva = cam.sva_nn_tx(nn,1,tx, t7, 1)
            print('tx= %1.2f  sx= %2.3f  vx= %2.3f  ax= %2.3f' % (tx, all_sva[0][nn], all_sva[1][nn], all_sva[2][nn]))
            print('           s7= %2.3f  v7= %2.3f  a7= %2.3f' % (cam.sva(tx)[0], cam.sva(tx)[1], cam.sva(tx)[2])) 
            input('Key In !')
            cam.graph_close()



    if menu == 12:  # 切断線txよりカム曲線を取得するための予備テスト
        t70=1; t71=1.5
        alf=1/4; rat=1; t0=0; t3=0.5; t1=t3*alf; t2=t3*(1-alf); t4=t3
        t5=t3*(rat*alf+1); t6=t70-(t5-t4); 
        nn = 10; dt = (t71-t70)/nn
        for i in range(nn):
            t7=t70+dt*i
            tt = [t0,t1,t2,t3,t4,t5,t6,t7]  
            cam = Universal_cam(0, 0.0 ,rat, *tt)
                           # nn, mode, scl_x, max_x 
            all_sva = cam.sva_nn(100,1,t7,1)
            print('t7= %1.2f' % t7)
            input('Key In !')
            cam.graph_close()



    if menu == 11:  # s/v指定変形台形カム曲線のためのテスト
        t0=0; t7=1; t30 = 0.3; t31 = 0.6; alf = 1/4
        nn = 10; dt = (t31-t30)/nn; rat = 1.2
        for i in range(nn):
            t3=t30+i*dt; t1=alf*t3; t2=t3*(1-alf); t4=t3
            t5=t3*(rat*alf+1); t6=t7-(t5-t4)
            tt = [t0,t1,t2,t3,t4,t5,t6,t7]  
            cam = Universal_cam(0, 0.0 ,rat, *tt)
            all_sva = cam.sva_nn(100,1,t7,3.0)
            print('s7= %2.3f  v7= %2.3f' % (cam.s7, cam.v7))
            input('Key In !')
            cam.graph_close()



    if menu == 10:   #zenkin(func, t00, t01, dt, target_val, gosa0)
        #m03 = 1/4; m47 = 1/4; rat = 2   #  t3 = 0.60128
        #m03 = 1/4; m47 = 3/4; rat = 2   #  NG!
        m03 = 1/4; m47 = 1/4; rat = 1.5  #  t3 = 0.54114
        #m03 = 1/4; m47 = 3/4; rat = 1.5  #  NG! 
        t0=0.0; t1=0.0; t6=1; t7=1

        def get_v3(t3):
            t2=(1-m03)*t3; t4=1-t3/rat; t5=t4+(1-t4)*m47 # 無停留変形等速度
            if t4 < t3:
                return (None, None) 
            else:
                tt = [0.0,t1,t2,t3,t4,t5,t6,1.0]  
                cam = Universal_cam(0, 0.0 ,rat, *tt)

                (v0, v1, v2, v3, v4, v5, v6, v7) = cam.get_v()
                #(ratio, a_mp, a_mm) = cam.get_ratio()
                #t_h = v3/vm * s_h   # ストローク時間[sec]
                #aa_mm = s_h/t_h**2*a_mm/grv     # 実加速度[G]

            return (cam ,v3)

        # func, t00, t01, dt, target_val, gosa0)
        cam = zenkin_cam(get_v3, 0.1, 0.7, 0.01, 1.8, 0.000001)
        #t2=(1-m03)*t3; t4=1-t3/rat; t5=t4+(1-t4)*m47
        #tt = [0.0,t1,t2,t3,t4,t5,t6,1.0]  # 速度指定型等速度曲線
        #cam = Universal_cam(0, 0.0 ,rat, *tt)
        if cam != None:
            all_sva0 = cam.sva_nn(100,1) 
            print('t3 = %1.5f' % cam.t3)
            input('Key In !')
        else:
            print('Cam is None!')



    if menu == 90:    # Cam No
        ss_cam_no, sxy = input('Key In Cam No & s,x,y=   ').split()
        cam_no = int(ss_cam_no)
        cam = Universal_cam(cam_no) 
        all_sva = cam.sva_nn_real(100,1,1,1,sxy) 
        #   (nn, mode, h, t_h, sxy = 's', s00=0)        


    if menu == 8:    # Cam No45 
        #cam = Universal_cam(45)             # v7 != 0 
        #cam = Universal_cam(45, 0.0 ,1.0)  # v7 != 0 
        cam = Universal_cam(45, 0.0 ,0.8)  # v7 != 0 
        all_sva = cam.sva_nn(100,1) 
        print('m= %2.5f     V6= %2.5f' % (cam.ratio, cam.v6))
        input('Key In!') 


    if menu == 7:   #zenkin(func, t00, t01, dt, target_val, gosa0)
        m03 = 1/4; m47 = 1/4; rat = 2   #  t3 = 0.60128
        #m03 = 1/4; m47 = 3/4; rat = 2   #  NG!
        #m03 = 1/4; m47 = 1/4; rat = 1.5  #  t3 = 0.54114
        #m03 = 1/4; m47 = 3/4; rat = 1.5  #  NG! 
        
        t0=0.0; t1=0.0; t6=1; t7=1

        def get_v3(t3):
            t2=(1-m03)*t3; t4=1-t3/rat; t5=t4+(1-t4)*m47 # 無停留変形等速度
            if t4 < t3:
                return 0.0 
            else:
                tt = [0.0,t1,t2,t3,t4,t5,t6,1.0]  
                cam = Universal_cam(0, 0.0 ,rat, *tt)

                (v0, v1, v2, v3, v4, v5, v6, v7) = cam.get_v()
                #(ratio, a_mp, a_mm) = cam.get_ratio()
                #t_h = v3/vm * s_h   # ストローク時間[sec]
                #aa_mm = s_h/t_h**2*a_mm/grv     # 実加速度[G]

            return v3

        # func, t00, t01, dt, target_val, gosa0)
        t3 = zenkin(get_v3, 0.1, 0.7, 0.01, 1.8, 0.000001)
        if t3 == None:
            print('いいのが無かったです')
            sys.exit() 

        t2=(1-m03)*t3; t4=1-t3/rat; t5=t4+(1-t4)*m47
        tt = [0.0,t1,t2,t3,t4,t5,t6,1.0]  # 速度指定型等速度曲線
        cam = Universal_cam(0, 0.0 ,rat, *tt)
        all_sva0 = cam.sva_nn(100,1) 
        print('t3 = %1.5f' % t3)
        input('Key In !')


    if menu == 6: # 無停留変形台形  
        t3 = 1/4; m03 = 1/4; rat = 2
        t0=0.0; t1=0.0; t2=(1-m03)*t3; t4=1-t3/rat; t5=t4+(1-t4)*m03; t6=1; t7=1
        print(t1,t2,t3,t4,t5)
        tt = (t0,t1,t2,t3,t4,t5,t6,t7)
        
        cam = Universal_cam(0, 0.0, rat, *tt)  #
        all_sva0 = cam.sva_nn(100,1)
        print(cam.v3)
        input('Key In !')

    elif menu == 5:  # 有次元、05T軸反転カム X軸反転カム、Y軸反転カム
        cam = Universal_cam(34, 0.2, 0.8)  #片停留変形台形(m=0.8)
        #cam = Universal_cam(45)  #片停留変形台形(m=1)
        all_svar = cam.sva_nn_real(100,1, 10.0, 3.0)        # 標準カム
        #input('Key In !')
        #cam.graph_close()
        # 反転cam
        cam_05Tm = cam.get_05Tmirror()
        all_sva = cam_05Tm.sva_nn_real(100,1,10.0,3.0)     # 0.5T軸反転カム
        #input('Key In !')
        #cam.graph_close()

        all_sva_xm = cam.sva_nn_real(100,1,10.0,3.0,'x')   # X軸反転カム
        #input('Key In !')
        #cam.graph_close()
        all_sva_ym = cam.sva_nn_real(100,1,10.0,3.0,'y')   ### Y軸反転カム
        #input('Key In !')
        #cam.graph_close()
       


    elif menu == 4:  # 標準変形正弦のグラフ作図
        cam26 = Universal_cam(cam_no = 26)  # 標準変形正弦
        all_sva26 = cam26.sva_nn(100,1)           # 100分割 plt on
        input('Key In !')
        #print(cam26.sva26(0.5))
        #print(all_sva26[0])          # 変位の全てのリスト   
        #print(all_sva26[0][1])       # 変位の2番目の値   
        #print(all_sva26[1])          # 速度の全てのリスト   
    
    elif menu == 3: # 早送り
        cam1 =  Universal_cam(cam_no = 1, z = 0.8)   # 早送り
        all_sva1 = cam1.sva_nn(100,1)
        input('Key In !')

    elif menu == 2: # 切削送り 
        cam2 =  Universal_cam(cam_no = 2)   # 切削送り 
        all_sva2 = cam2.sva_nn(100,1)  
        input('Key In !')

    elif menu == 9: # s -> t 逆算
        cam26 = Universal_cam(cam_no = 26)  # 標準変形正弦
        target_s = 0.2; gosa = 0.00001
        target_t = cam26.s2t(target_s, gosa)

        print('target_s= %1.2f  starget_t= %1.8f' % (target_s,target_t))
        print(cam26.ttl)
        input('Key In !')
